@@ -1,15 +1,47 @@
 /*
-** This is from the NIST implementation of  FIPS-181, so I CAN NOT
-** copyright it.
-** It is modified to compile itself
+** This module uses code from the NIST implementation of  FIPS-181,
+** but the algorythm is CHANGED and I think that I CAN
+** copyright it. See copiright notes below.
 */
 
+/*
+** Copyright (c) 1999, 2000, 2001
+** Adel I. Mirzazhanov. All rights reserved
+**
+** Redistribution and use in source and binary forms, with or without
+** modification, are permitted provided that the following conditions
+** are met:
+** 
+**     1.Redistributions of source code must retain the above copyright notice,
+**       this list of conditions and the following disclaimer. 
+**     2.Redistributions in binary form must reproduce the above copyright
+**       notice, this list of conditions and the following disclaimer in the
+**       documentation and/or other materials provided with the distribution. 
+**     3.The name of the author may not be used to endorse or promote products
+**       derived from this software without specific prior written permission. 
+** 		  
+** THIS SOFTWARE IS PROVIDED BY THE AUTHOR  ``AS IS'' AND ANY EXPRESS
+** OR IMPLIED WARRANTIES, INCLUDING,  BUT NOT LIMITED TO, THE IMPLIED
+** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+** ARE DISCLAIMED.  IN  NO  EVENT  SHALL THE AUTHOR BE LIABLE FOR ANY
+** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+** DAMAGES (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE
+** GOODS OR SERVICES;  LOSS OF USE,  DATA,  OR  PROFITS;  OR BUSINESS
+** INTERRUPTION)  HOWEVER  CAUSED  AND  ON  ANY  THEORY OF LIABILITY,
+** WHETHER  IN  CONTRACT,   STRICT   LIABILITY,  OR  TORT  (INCLUDING
+** NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+*/
+
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h> 
 #include <time.h>
 #include <sys/types.h>
 #include "pronpass.h"
+#include "randpass.h"
 #include "errs.h"
 
 struct unit
@@ -1232,8 +1264,9 @@ static int  digram[][RULE_SIZE] =
 */
 int
 gen_pron_pass (char *word, char *hyphenated_word, USHORT minlen,
-               USHORT maxlen)
+               USHORT maxlen, unsigned int pass_mode)
 {
+
     int     pwlen;
 
  /* 
@@ -1256,7 +1289,7 @@ gen_pron_pass (char *word, char *hyphenated_word, USHORT minlen,
  /* 
   * Find password.
   */
-    pwlen = gen_word (word, hyphenated_word, get_random (minlen, maxlen));
+    pwlen = gen_word (word, hyphenated_word, get_random (minlen, maxlen), pass_mode);
     return (pwlen);
 }
 
@@ -1271,7 +1304,7 @@ gen_pron_pass (char *word, char *hyphenated_word, USHORT minlen,
  * will be found if the retry limit is adequately large enough.
  */
 int
-gen_word (char *word, char *hyphenated_word, USHORT pwlen)
+gen_word (char *word, char *hyphenated_word, USHORT pwlen, unsigned int pass_mode)
 {
     USHORT word_length;
     USHORT syllable_length;
@@ -1282,6 +1315,7 @@ gen_word (char *word, char *hyphenated_word, USHORT pwlen)
     USHORT *word_units;
     USHORT syllable_size;
     UINT   tries;
+    int ch_flag = FALSE;
 
     /*
      * Keep count of retries.
@@ -1309,9 +1343,9 @@ gen_word (char *word, char *hyphenated_word, USHORT pwlen)
      * explicit rule limits the length of syllables, but digram rules and
      * heuristics do so indirectly.
      */
-    if ( (word_units     = (USHORT *) calloc (sizeof (USHORT), pwlen))==NULL ||
-         (syllable_units = (USHORT *) calloc (sizeof (USHORT), pwlen))==NULL ||
-         (new_syllable   = (char *) calloc (sizeof (USHORT), pwlen))  ==NULL)
+    if ( (word_units     = (USHORT *) calloc (sizeof (USHORT), pwlen+1))==NULL ||
+         (syllable_units = (USHORT *) calloc (sizeof (USHORT), pwlen+1))==NULL ||
+         (new_syllable   = (char *) calloc (sizeof (USHORT), pwlen+1))  ==NULL)
 	   return(-1);
 
     /*
@@ -1324,7 +1358,7 @@ gen_word (char *word, char *hyphenated_word, USHORT pwlen)
       */
      (void) gen_syllable (new_syllable, pwlen - word_length, syllable_units, &syllable_size);
      syllable_length = strlen (new_syllable);
-
+     
      /*
       * Append the syllable units to the word units.
       */
@@ -1349,14 +1383,42 @@ gen_word (char *word, char *hyphenated_word, USHORT pwlen)
      {
          if (word_length == 0)
          {
+          /*
+          ** Modify syllable for numeric or capital symbols required
+          ** Should be done after word quality check. 
+          */
+          if ( ((pass_mode & S_NB) > 0) && (syllable_length == 1))
+            {
+             numerize(new_syllable);
+	     ch_flag = TRUE;
+            }
+          if ( ( (pass_mode & S_CL) > 0) && (ch_flag != TRUE))
+             capitalize(new_syllable);
+          ch_flag = FALSE;
+          /**/
           (void) strcpy (word, new_syllable);
           (void) strcpy (hyphenated_word, new_syllable);
+	  bzero ( (void *)new_syllable, (size_t)(pwlen * sizeof(USHORT)+1));
          }
          else
          {
+          /*
+          ** Modify syllable for numeric or capital symbols required
+          ** Should be done after word quality check.
+          */
+          if ( ((pass_mode & S_NB) > 0) && (syllable_length == 1))
+            {
+             numerize(new_syllable);
+	     ch_flag = TRUE;
+            }
+          if ( ( (pass_mode & S_CL) > 0) && (ch_flag != TRUE))
+             capitalize(new_syllable);
+          ch_flag = FALSE;
+          /**/
           (void) strcat (word, new_syllable);
           (void) strcat (hyphenated_word, "-");
           (void) strcat (hyphenated_word, new_syllable);
+	  bzero ( (void *)new_syllable, (size_t)(pwlen * sizeof(USHORT)+1));
          }
          word_length += syllable_length;
      }
@@ -1565,18 +1627,18 @@ char *
 gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
               USHORT *syllable_length)
 {
-    USHORT unit;
-    SHORT  current_unit;
-    USHORT vowel_count;
+    USHORT  unit = 0;
+    SHORT   current_unit = 0;
+    USHORT  vowel_count = 0;
     boolean rule_broken;
     boolean want_vowel;
     boolean want_another_unit;
-    UINT	tries;
-    USHORT  	last_unit;
-    SHORT	 length_left;
-    USHORT 	hold_saved_unit;
-    static 	USHORT saved_unit;
-    static 	USHORT saved_pair[2];
+    UINT    tries = 0;
+    USHORT  last_unit = 0;
+    SHORT   length_left = 0;
+    USHORT  hold_saved_unit = 0;
+    static  USHORT saved_unit;
+    static  USHORT saved_pair[2];
 
     /*
      * This is needed if the saved_unit is tries and the syllable then
@@ -1662,7 +1724,6 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                unit = random_unit (VOWEL);
               else
                unit = random_unit (NO_SPECIAL_RULE);
-
           length_left -= (short int) strlen (rules[unit].unit_code);
 
           /*
@@ -1695,10 +1756,12 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                 * discard it.
                 */
                if (length_left == 0)
+	          {
                    if (rules[unit].flags & VOWEL)
                     want_another_unit = FALSE;
                    else
                     rule_broken = TRUE;
+		  }
           }
           else
           {
@@ -1795,7 +1858,7 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                     ((length_left > 0) ||
                         !(rules[last_unit].flags &
                          NO_FINAL_SPLIT)))
-
+                  {
                    /*
                     * Don't allow 3 consecutive vowels in a
                     * syllable.  Although some words formed like this
@@ -1834,6 +1897,7 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                          want_another_unit = FALSE;
                         }
                     }
+		  }
               }
 
               /*
@@ -1848,11 +1912,10 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                /*
                 * This word ends in a silent e.
                 */
-               if (((vowel_count != 0) &&
-                        (rules[unit].flags & NO_FINAL_SPLIT) &&
-                        (length_left == 0) &&
-                        !(rules[last_unit].flags &
-                         VOWEL)) ||
+/******/        if (((vowel_count != 0) &&
+                     (rules[unit].flags & NO_FINAL_SPLIT) &&
+                     (length_left == 0) &&
+                    !(rules[last_unit].flags & VOWEL)) ||
 
                     /*
                      * This syllable ends either because the digram
@@ -1860,8 +1923,10 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                      * the length of the word.
                      */
                     (ALLOWED (END) || (length_left == 0)))
+		   {
                    want_another_unit = FALSE;
-               else
+		   }
+	       else
                    /*
                     * Since we have a vowel in the syllable
                     * already, if the digram calls for the end of the
@@ -1883,8 +1948,7 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                     if (ALLOWED (BEGIN) &&
                          (current_unit > 1) &&
                          !((vowel_count == 1) &&
-                             (rules[last_unit].flags &
-                              VOWEL)))
+                         (rules[last_unit].flags & VOWEL)))
                     {
                         saved_unit = 2;
                         saved_pair[0] = unit;
@@ -1901,10 +1965,12 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
                    }
                    else
                     if (ALLOWED (SUFFIX))
+		     {
                         want_vowel = TRUE;
+		     }
               }
           }
-
+/********/
           tries++;
 
           /*
@@ -1989,9 +2055,9 @@ gen_syllable (char *syllable, USHORT pwlen, USHORT *units_in_syllable,
 boolean
 illegal_placement (USHORT *units, USHORT pwlen)
 {
-     USHORT vowel_count;
-     USHORT unit_count;
-        boolean failure;
+    USHORT vowel_count;
+    USHORT unit_count;
+    boolean failure;
 
     vowel_count = 0;
     failure = FALSE;
@@ -2151,12 +2217,12 @@ random_unit (USHORT type)
      * to do that and avoid looping with rejected consonants.
      */
     if (type & VOWEL)
-     number = vowel_numbers[get_random (0, sizeof (vowel_numbers) / sizeof (USHORT))];
+     number = vowel_numbers[get_random (0, (sizeof (vowel_numbers) / sizeof (USHORT))-1)];
     else
      /* 
       * Get any letter according to the English distribution.
       */
-     number = numbers[get_random (0, sizeof (numbers) / sizeof (USHORT))];
+     number = numbers[get_random (0, (sizeof (numbers) / sizeof (USHORT))-1)];
     return (number);
 }
 
@@ -2172,5 +2238,56 @@ random_unit (USHORT type)
 USHORT
 get_random (USHORT minlen, USHORT maxlen)
 {
- return minlen + (USHORT) randint ((int) (maxlen - minlen + 1));
+ USHORT ret = 0;
+ ret = minlen + (USHORT) randint ((int) (maxlen - minlen + 1));
+ return (ret);
+}
+
+/*
+** This routine designed to modify sullable like this:
+** adel ----> Adel
+** dot  ----> Dot
+** etc.
+*/
+void capitalize (char *syllable)
+{
+ char let[26] =
+ {
+ 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
+ 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
+ 'u', 'v', 'w', 'x', 'w', 'z'
+ };
+ char clet[26] =
+ {
+ 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
+ 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
+ 'U', 'V', 'W', 'X', 'W', 'Z'
+  };
+ char tmp = 0x00;
+ int i = 0;
+ if ( randint(2) == TRUE)
+  {
+   bcopy ((void *)syllable, (void *)&tmp, sizeof(tmp));
+   for(i=0; i < 26; i++)
+    if ( let[i] == tmp ) bcopy ((void *)&clet[i], (void *)syllable, 1);
+  }
+}
+
+/*
+** This routine designed to modify single-letter syllable like this:
+** a ----> 1 or 2 or 3 etc.
+** u ----> 1 or 2 or 3 etc.
+** etc.
+*/
+void numerize (char *syllable)
+{
+ char *tmp;
+ if ( (tmp = (char *)calloc(1, 4)) == NULL)
+    err_sys_fatal("calloc");
+ if ( (randint(2) == TRUE) && (strlen (syllable) == 1) )
+      {
+       sprintf(tmp, "%d", randint(10));
+       bcopy ((void *)tmp, (void *)syllable, 1);
+      }
+ free ((void *)tmp);
 }
